from nomenklatura import store as nk
from nomenklatura.store import redis

# from ftmq.model import Catalog
from ftmq.store import CE, DS, Store, View


class RedisQueryView(View, redis.RedisView):
    pass


class RedisStore(Store, redis.RedisStore):
    def query(self, scope: DS | None = None, external: bool = False) -> nk.View[DS, CE]:
        scope = scope or self.dataset
        return RedisQueryView(self, scope, external=external)



import re
from functools import cache, lru_cache
from typing import Any, Generator

import pycountry
from banal import ensure_list, is_listish
from followthemoney.proxy import E, EntityProxy
from followthemoney.schema import Schema
from followthemoney.types import registry
from followthemoney.util import make_entity_id, sanitize_text
from nomenklatura.dataset import Dataset
from nomenklatura.entity import CE, CompositeEntity
from nomenklatura.statement import Statement
from nomenklatura.stream import StreamEntity
from normality import collapse_spaces, slugify

from ftmq.enums import Comparators
from ftmq.exceptions import ValidationError
from ftmq.types import SE, SGenerator


@cache
def make_dataset(name: str) -> Dataset:
    return Dataset.make({"name": name, "title": name})


@cache
def ensure_dataset(ds: str | Dataset | None = None) -> Dataset | None:
    if not ds:
        return
    if isinstance(ds, str):
        return make_dataset(ds)
    return ds


DefaultDataset = make_dataset("default")


def parse_comparator(key: str) -> tuple[str, Comparators]:
    key, *comparator = key.split("__", 1)
    if comparator:
        comparator = Comparators[comparator[0]]
    else:
        comparator = Comparators["eq"]
    return key, comparator


def parse_unknown_filters(
    filters: tuple[str],
) -> Generator[tuple[str, str, str], None, None]:
    filters = (f for f in filters)
    for prop in filters:
        prop = prop.lstrip("-")
        if "=" in prop:  # 'country=de'
            prop, value = prop.split("=")
        else:  # ("country", "de"):
            value = next(filters)

        prop, *op = prop.split("__")
        op = op[0] if op else Comparators.eq
        if op == Comparators["in"]:
            # de,fr or ["de", "fr"]
            if is_listish(value):
                value = ensure_list(value)
            else:
                value = value.split(",")
        yield prop, value, op


def make_proxy(data: dict[str, Any], dataset: str | Dataset | None = None) -> CE:
    """
    Create a `nomenklatura.entity.CompositeEntity` from a json dict.

    Args:
        data: followthemoney data dict that represents entity data.
        dataset: A default dataset

    Returns:
        The composite entity proxy
    """
    datasets = ensure_list(data.pop("datasets", None))
    if dataset is not None:
        if isinstance(dataset, str):
            dataset = make_dataset(dataset)
        datasets.append(dataset.name)
    elif datasets:
        dataset = datasets[0]
        dataset = make_dataset(dataset)
    else:
        dataset = DefaultDataset
    proxy = CompositeEntity(dataset, data)
    if len(datasets) > 1:
        if proxy.id is None:
            raise ValidationError("Entity has no ID.")
        statements = get_statements(proxy, *datasets)
        return CompositeEntity.from_statements(dataset, statements)
    return proxy


def ensure_proxy(
    data: dict[str, Any] | CE | E | SE, dataset: str | Dataset | None = None
) -> CompositeEntity:
    if isinstance(data, CompositeEntity):
        return data
    if isinstance(data, (EntityProxy, StreamEntity)):
        data = data.to_full_dict()
    return make_proxy(data, dataset)


def get_statements(proxy: CE, *datasets: str) -> SGenerator:
    """
    Get statements from a `nomenklatura.entity.CompositeEntity` with multiple
    datasets if needed

    Args:
        proxy: `nomenklatura.entity.CompositeEntity`
        *datasets: Any (additional) datasets to create statements for

    Yields:
        A generator of `nomenklatura.statement.Statement`
    """
    datasets = datasets or ("default",)
    for ix, dataset in enumerate(datasets):
        for sx, stmt in enumerate(Statement.from_entity(proxy, dataset)):
            stmt = stmt.to_dict()
            stmt["external"] = stmt.get("external") or False
            stmt = Statement.from_dict(stmt)
            yield stmt
            if ix and sx:
                break


@cache
def get_country_name(code: str) -> str:
    """
    Get the (english) country name for the given 2-letter iso code via
    [pycountry](https://pypi.org/project/pycountry/)

    Examples:
        >>> get_country_name("de")
        "Germany"
        >>> get_country_name("xx")
        "xx"
        >>> get_country_name("gb") == get_country_name("uk")
        True  # United Kingdom

    Args:
        code: Two-letter iso code, case insensitive

    Returns:
        Either the country name for a valid code or the code as fallback.
    """
    code_clean = get_country_code(code)
    if code_clean is None:
        code_clean = code.lower()
    try:
        country = pycountry.countries.get(alpha_2=code_clean)
        if country is not None:
            return country.name
    except (LookupError, AttributeError):
        return code
    return code_clean


@lru_cache(1024)
def get_country_code(value: Any, splitter: str | None = ",") -> str | None:
    """
    Get the 2-letter iso country code for an arbitrary country name

    Examples:
        >>> get_country_code("Germany")
        "de"
        >>> get_country_code("Deutschland")
        "de"
        >>> get_country_code("Berlin, Deutschland")
        "de"
        >>> get_country_code("Foo")
        None

    Args:
        value: Any input that will be [cleaned][ftmq.util.clean_string]
        splitter: Character to use to get text tokens to find country name for

    Returns:
        The iso code or `None`
    """
    value = clean_string(value)
    if not value:
        return
    code = registry.country.clean_text(value)
    if code:
        return code
    for token in value.split(splitter):
        code = registry.country.clean_text(token)
        if code:
            return code
    return


NUMERIC_US = re.compile(r"^-?\d+(?:,\d{3})*(?:\.\d+)?$")
NUMERIC_DE = re.compile(r"^-?\d+(?:\.\d{3})*(?:,\d+)?$")


def to_numeric(value: str) -> float | int | None:
    """
    Convert a string value into a primitive numeric dtype (`int` or `float`)
    taking US and DE formatting into account via regex

    Examples:
        >>> to_numeric("1")
        1
        >>> to_numeric("1.0")
        1
        >>> to_numeric("1.1")
        1.1
        >>> to_numeric("1,101,000")
        1101000
        >>> to_numeric("1.000,1")
        1000.1
        >>> to_numeric("foo")
        None

    Args:
        value: The input

    Returns:
        The converted number or `None` if conversion fails
    """
    value = str(value).strip()
    try:
        value = float(value)
        if int(value) == value:
            return int(value)
        return value
    except ValueError:
        if re.match(NUMERIC_US, value):
            return to_numeric(value.replace(",", ""))
        if re.match(NUMERIC_DE, value):
            return to_numeric(value.replace(".", "").replace(",", "."))


def join_slug(
    *parts: str | None,
    prefix: str | None = None,
    sep: str = "-",
    strict: bool = True,
    max_len: int = 255,
) -> str | None:
    """
    Create a stable slug from parts with optional validation

    Examples:
        >>> join_slug("foo", "bar")
        "foo-bar"
        >>> join_slug("foo", None, "bar")
        None
        >>> join_slug("foo", None, "bar", strict=False)
        "foo-bar"
        >>> join_slug("foo", "bar", sep="_")
        "foo_bar"
        >>> join_slug("a very long thing", max_len=15)
        "a-very-5c156cf9"

    Args:
        *parts: Multiple (ordered) parts to compute the slug from
        prefix: Add a prefix to the slug
        sep: Parts separator
        strict: Ensure all parts are not `None`
        max_len: Maximum length of the slug. If it exceeds, the returned value
            will get a computed hash suffix

    Returns:
        The computed slug or `None` if validation fails
    """
    sections = [slugify(p, sep=sep) for p in parts]
    if strict and None in sections:
        return None
    texts = [p for p in sections if p is not None]
    if not len(texts):
        return None
    prefix = slugify(prefix, sep=sep)
    if prefix is not None:
        texts = [prefix, *texts]
    slug = sep.join(texts)
    if len(slug) <= max_len:
        return slug
    # shorten slug but ensure uniqueness
    ident = make_entity_id(slug)[:8]
    slug = slug[: max_len - 9].strip(sep)
    return f"{slug}-{ident}"


def get_year_from_iso(value: Any) -> int | None:
    """
    Extract the year from a iso date string or `datetime` object.

    Examples:
        >>>  get_year_from_iso(None)
        None
        >>>  get_year_from_iso("2023")
        2023
        >>>  get_year_from_iso(2020)
        2020
        >>>  get_year_from_iso(datetime.now())
        2024
        >>>  get_year_from_iso("2000-01")
        2000

    Args:
        value: Any input that will be [cleaned][ftmq.util.clean_string]

    Returns:
        The year or `None`
    """
    value = clean_string(value)
    if not value:
        return
    try:
        return int(str(value)[:4])
    except ValueError:
        return


@lru_cache(1024)
def clean_string(value: Any) -> str | None:
    """
    Convert a value to `None` or a sanitized string without linebreaks

    Examples:
        >>> clean_string(" foo\n bar")
        "foo bar"
        >>> clean_string("foo Bar, baz")
        "foo Bar, baz"
        >>> clean_string(None)
        None
        >>> clean_string("")
        None
        >>> clean_string("  ")
        None
        >>> clean_string(100)
        "100"

    Args:
        value: Any input that will be converted to string

    Returns:
        The cleaned value or `None`
    """
    value = sanitize_text(value)
    if value is None:
        return
    return collapse_spaces(value)


@lru_cache(1024)
def clean_name(value: Any) -> str | None:
    """
    Clean a value and only return it if it is a "name" in the sense of, doesn't
    contain exclusively of special chars

    Examples:
        >>> clean_name("  foo\n Bar")
        "foo Bar"
        >>> clean_name("- - . *")
        None

    Args:
        value: Any input that will be [cleaned][ftmq.util.clean_string]

    Returns:
        The cleaned name or `None`
    """
    value = clean_string(value)
    if slugify(value) is None:
        return
    return value


@lru_cache(1024)
def make_fingerprint(value: Any) -> str | None:
    """
    Create a stable but simplified string or `None` from input that can be used
    to generate ids (to mimic `fingerprints.generate` which is unstable for IDs
    as its algorithm could change)

    Examples:
        >>> make_fingerprint("Mrs. Jane Doe")
        "doe jane mrs"
        >>> make_fingerprint("Mrs. Jane Mrs. Doe")
        "doe jane mrs"
        >>> make_fingerprint("#")
        None
        >>> make_fingerprint(" ")
        None
        >>> make_fingerprint("")
        None
        >>> make_fingerprint(None)
        None

    Args:
        value: Any input that will be [cleaned][ftmq.util.clean_name]

    Returns:
        The simplified string (fingerprint) or `None` if value is not feasible
            to fingerprint.
    """
    value = clean_name(value)
    if value is None:
        return
    return " ".join(sorted(set(slugify(value).split("-"))))


@lru_cache(1024)
def make_string_id(*values: Any) -> str | None:
    """
    Compute a hash id based on values

    Args:
        *values: Parts to compute id from that will be
            [cleaned][ftmq.util.clean_name]

    Returns:
        The computed hash id or `None` if a parts cleaned value is `None`
    """
    return make_entity_id(*map(clean_name, values))


@lru_cache(1024)
def make_fingerprint_id(*values: Any) -> str | None:
    """
    Compute a hash id based on values fingerprints

    Args:
        *values: Parts to compute id from that will be
            [fingerprinted][ftmq.util.make_fingerprint]

    Returns:
        The computed hash id or `None` if a parts fingerprinted value is `None`
    """
    return make_entity_id(*map(make_fingerprint, values))


@cache
def prop_is_numeric(schema: Schema, prop: str) -> bool:
    """
    Indicate if the given property is numeric type

    Args:
        schema: followthemoney schema
        prop: Property

    Returns:
        `False` if the property is not numeric type or not found in the schema
            at all
    """
    prop_ = schema.get(prop)
    if prop_ is not None:
        return prop_.type == registry.number
    return False


def get_proxy_caption_property(proxy: CE) -> dict[str, str]:
    for prop in proxy.schema.caption:
        for value in proxy.get(prop):
            return {prop: value}
    return {}


def get_dehydrated_proxy(proxy: CE) -> CE:
    """
    Reduce proxy payload to only include caption property

    Args:
        proxy: `nomenklatura.entity.CompositeEntity`

    Returns:
        A `nomenklatura.entity.CompositeEntity` with only the caption property.
    """
    return make_proxy(
        {
            "id": proxy.id,
            "schema": proxy.schema.name,
            "properties": get_proxy_caption_property(proxy),
            "datasets": proxy.datasets,
        }
    )


def get_featured_proxy(proxy: CE) -> CE:
    """
    Reduce proxy payload to only include featured properties

    Args:
        proxy: `nomenklatura.entity.CompositeEntity`

    Returns:
        A `nomenklatura.entity.CompositeEntity` with only the featured
            properties for its schema.
    """
    featured = get_dehydrated_proxy(proxy)
    for prop in proxy.schema.featured:
        featured.add(prop, proxy.get(prop))
    return featured


def must_str(value: Any) -> str:
    value = clean_string(value)
    if not value:
        raise ValueError(f"Value invalid: `{value}`")
    return value



import os
from collections.abc import Generator
from pathlib import Path
from typing import Literal, TypeAlias, TypeVar

from anystore.types import BytesGenerator, SDict, SDictGenerator, StrGenerator
from nomenklatura.entity import CE
from nomenklatura.statement.statement import Statement
from nomenklatura.stream import StreamEntity

from ftmq import enums

# property multi-value
Value: TypeAlias = list[str]

SE = TypeVar("SE", bound=StreamEntity)
Proxy: TypeAlias = SE | CE

# entity generators
CEGenerator: TypeAlias = Generator[CE, None, None]
SEGenerator: TypeAlias = Generator[SE, None, None]
ProxyGenerator: TypeAlias = Generator[Proxy, None, None]

# statement generator
SGenerator: TypeAlias = Generator[Statement, None, None]

Schemata: TypeAlias = Literal[tuple(s.name for s in enums.Schemata)]
Properties: TypeAlias = Literal[tuple(p.name for p in enums.Properties)]
Frequencies: TypeAlias = Literal[tuple(f.name for f in enums.Frequencies)]

PathLike: TypeAlias = str | os.PathLike[str] | Path

__all__ = [
    "BytesGenerator",
    "CE",
    "SE",
    "Proxy",
    "CEGenerator",
    "SEGenerator",
    "ProxyGenerator",
    "Frequencies",
    "PathLike",
    "Properties",
    "Schemata",
    "SDict",
    "SGenerator",
    "SDictGenerator",
    "StrGenerator",
    "Value",
]



from typing import Any, Iterable

import orjson
from anystore.io import Uri, smart_open, smart_stream
from banal import is_listish
from followthemoney import model
from nomenklatura.entity import CE, CompositeEntity
from nomenklatura.stream import StreamEntity

from ftmq.logging import get_logger
from ftmq.query import Query
from ftmq.store import Store, get_store
from ftmq.types import CEGenerator, Proxy, SEGenerator
from ftmq.util import ensure_proxy, get_statements, make_dataset, make_proxy

log = get_logger(__name__)

DEFAULT_MODE = "rb"


def smart_get_store(uri: Uri, **kwargs) -> Store | None:
    try:
        return get_store(uri, **kwargs)
    except NotImplementedError:
        return


def smart_read_proxies(
    uri: Uri | Iterable[Uri],
    mode: str | None = DEFAULT_MODE,
    query: Query | None = None,
    **store_kwargs: Any,
) -> CEGenerator:
    """
    Stream proxies from an arbitrary source

    Example:
        ```python
        from ftmq import Query
        from ftmq.io import smart_read_proxies

        # remote file-like source
        for proxy in smart_read_proxies("s3://data/entities.ftm.json"):
            print(proxy.schema)

        # multiple files
        for proxy in smart_read_proxies("./1.json", "./2.json"):
            print(proxy.schema)

        # nomenklatura store
        for proxy in smart_read_proxies("redis://localhost", dataset="default"):
            print(proxy.schema)

        # apply a query to sql storage
        q = Query(dataset="my_dataset", schema="Person")
        for proxy in smart_read_proxies("sqlite:///data/ftm.db", query=q):
            print(proxy.schema)
        ```

    Args:
        uri: File-like uri or store uri or multiple uris
        mode: Open mode for file-like sources (default: `rb`)
        query: Filter `Query` object
        **store_kwargs: Pass through configuration to statement store

    Yields:
        A stream of `nomenklatura.entity.CompositeEntity`
    """
    if is_listish(uri):
        for u in uri:
            yield from smart_read_proxies(u, mode, query)
        return

    store = smart_get_store(uri, **store_kwargs)
    if store is not None:
        view = store.query()
        yield from view.entities(query)
        return

    q = query or Query()
    lines = smart_stream(uri)
    lines = (orjson.loads(line) for line in lines)
    proxies = (make_proxy(line) for line in lines)
    yield from q.apply_iter(proxies)


def smart_stream_proxies(
    uri: Uri | Iterable[Uri], mode: str | None = DEFAULT_MODE
) -> SEGenerator:
    """
    Stream `nomenklatura.stream.StreamEntity` from fs-like uris.

    Args:
        uri: File-like uri or multiple uris
        mode: Open mode for file-like sources (default: `rb`)

    Yields:
        A stream of `nomenklatura.stream.StreamEntity`
    """
    if is_listish(uri):
        for u in uri:
            yield from smart_stream_proxies(u, mode)
        return

    for line in smart_stream(uri, mode):
        data = orjson.loads(line)
        yield StreamEntity.from_dict(model, data)


def smart_write_proxies(
    uri: Uri,
    proxies: Iterable[Proxy],
    mode: str | None = "wb",
    **store_kwargs: Any,
) -> int:
    """
    Write a stream of proxies (or data dicts) to an arbitrary target.

    Example:
        ```python
        from ftmq.io import smart_write_proxies

        proxies = [...]

        # to a remote cloud storage
        smart_write_proxies("s3://data/entities.ftm.json", proxies)

        # to a redis statement store
        smart_write_proxies("redis://localhost", proxies, dataset="my_dataset")
        ```

    Args:
        uri: File-like uri or store uri
        proxies: Iterable of proxy data
        mode: Open mode for file-like targets (default: `wb`)
        **store_kwargs: Pass through configuration to statement store

    Returns:
        Number of written proxies
    """
    ix = 0
    if not proxies:
        return ix

    store = smart_get_store(uri, **store_kwargs)
    if store is not None:
        proxies = (ensure_proxy(p) for p in proxies)
        dataset = store_kwargs.get("dataset")
        if dataset is not None:
            proxies = apply_datasets(proxies, dataset, replace=True)
        with store.writer() as bulk:
            for proxy in proxies:
                ix += 1
                bulk.add_entity(proxy)
                if ix % 1_000 == 0:
                    log.info("Writing proxy %d ..." % ix)
        return ix

    with smart_open(uri, mode=mode) as fh:
        for proxy in proxies:
            ix += 1
            data = proxy.to_dict()
            fh.write(orjson.dumps(data, option=orjson.OPT_APPEND_NEWLINE))
    return ix


def apply_datasets(
    proxies: Iterable[CE], *datasets: Iterable[str], replace: bool | None = False
) -> CEGenerator:
    """
    Apply datasets to a stream of proxies

    Args:
        proxies: Iterable of `nomenklatura.entity.CompositeEntity`
        *datasets: One or more dataset names to apply
        replace: Drop any other existing datasets

    Yields:
        The proxy stream with the datasets applied
    """
    for proxy in proxies:
        if datasets:
            if not replace:
                datasets = proxy.datasets | set(datasets)
            statements = get_statements(proxy, *datasets)
            dataset = make_dataset(list(datasets)[0])
        yield CompositeEntity.from_statements(dataset, statements)



from collections.abc import Iterable
from itertools import islice
from typing import Any, TypeVar

from banal import ensure_list, is_listish, is_mapping
from nomenklatura.entity import CE

from ftmq.aggregations import Aggregation, Aggregator
from ftmq.enums import Aggregations, Properties
from ftmq.exceptions import ValidationError
from ftmq.filters import (
    FILTERS,
    DatasetFilter,
    F,
    IdFilter,
    PropertyFilter,
    ReverseFilter,
    SchemaFilter,
)
from ftmq.sql import Sql
from ftmq.types import CEGenerator
from ftmq.util import (
    parse_comparator,
    parse_unknown_filters,
    prop_is_numeric,
    to_numeric,
)

Q = TypeVar("Q", bound="Query")
Slice = TypeVar("Slice", bound=slice)


class Sort:
    def __init__(self, values: Iterable[str], ascending: bool | None = True) -> None:
        self.values = tuple(values)
        self.ascending = ascending

    def apply(self, proxy: CE) -> tuple[str]:
        values = tuple()
        for v in self.values:
            p_values = proxy.get(v, quiet=True) or []
            if prop_is_numeric(proxy.schema, v):
                p_values = map(to_numeric, p_values)
            values = values + (tuple(p_values))
        return values

    def apply_iter(self, proxies: CEGenerator) -> CEGenerator:
        yield from sorted(
            proxies, key=lambda x: self.apply(x), reverse=not self.ascending
        )

    def serialize(self) -> list[str]:
        if self.ascending:
            return list(self.values)
        return [f"-{v}" for v in self.values]


class Query:
    def __init__(
        self,
        filters: Iterable[F] | None = None,
        aggregations: Iterable[Aggregation] | None = None,
        aggregator: Aggregator | None = None,
        sort: Sort | None = None,
        slice: Slice | None = None,
    ):
        self.filters = set(ensure_list(filters))
        self.aggregations = set(ensure_list(aggregations))
        self.aggregator = aggregator
        self.sort = sort
        self.slice = slice

    def __getitem__(self, value: Any) -> Q:
        """
        Implement list-like slicing. No negative values allowed.

        Examples:
            >>> q[1]
            # 2nd element (0-index)
            >>> q[:10]
            # first 10 elements
            >>> q[10:20]
            # next 10 elements

        Returns:
            The updated `Query` instance
        """
        if isinstance(value, int):
            if value < 0:
                raise ValidationError(f"Invalid slicing: `{value}`")
            return self._chain(slice=slice(value, value + 1))
        if isinstance(value, slice):
            if value.step is not None:
                raise ValidationError(f"Invalid slicing: `{value}`")
            return self._chain(slice=value)
        raise NotImplementedError

    def __bool__(self) -> bool:
        """
        Detect if any filter, ordering or slicing is defined

        Examples:
            >>> bool(Query())
            False
            >>> bool(Query().where(dataset="my_dataset"))
            True
        """
        return bool(self.to_dict())

    def __hash__(self) -> int:
        """
        Generate a unique key of the current state, useful for caching
        """
        return hash(repr(self.to_dict()))

    def _chain(self, **kwargs):
        # merge current state
        new_kwargs = self.__dict__.copy()
        for key, new_value in kwargs.items():
            old_value = new_kwargs[key]
            if old_value is None:
                new_kwargs[key] = new_value
            # "remove" old value:
            elif new_value is None:
                new_kwargs[key] = None
            # overwrite order by
            elif key == "sort":
                new_kwargs[key] = new_value
            # combine iterables and dicts
            elif is_listish(old_value):
                new_kwargs[key] = sorted(set(old_value) | set(new_value))
            elif is_mapping(old_value):
                new_kwargs[key] = {**old_value, **new_value}
            else:  # replace
                new_kwargs[key] = new_value
        return self.__class__(**new_kwargs)

    def _get_lookups(self, filters: set[F]) -> dict[str, Any]:
        data = {}
        for fi in filters:
            for k, v in fi.to_dict().items():
                current = data.get(k)
                if is_listish(current):
                    data[k].append(v)
                else:
                    data[k] = v
        return data

    @property
    def lookups(self) -> dict[str, Any]:
        """
        The current filter lookups as dictionary
        """
        return self._get_lookups(self.filters)

    @property
    def limit(self) -> int | None:
        """
        The current limit (inferred from a slice)
        """
        if self.slice is None:
            return None
        if self.slice.start and self.slice.stop:
            return self.slice.stop - self.slice.start
        return self.slice.stop

    @property
    def offset(self) -> int | None:
        """
        The current offset (inferred from a slice)
        """
        return self.slice.start if self.slice else None

    @property
    def sql(self) -> Sql:
        """
        An object of this query used for sql interfaces
        """
        return Sql(self)

    @property
    def ids(self) -> set[IdFilter]:
        """
        The current id filters
        """
        return {f for f in self.filters if isinstance(f, IdFilter)}

    @property
    def datasets(self) -> set[DatasetFilter]:
        """
        The current dataset filters
        """
        return {f for f in self.filters if isinstance(f, DatasetFilter)}

    @property
    def dataset_names(self) -> set[str]:
        """
        The names of the current filtered datasets
        """
        names = set()
        for f in self.datasets:
            names.update(ensure_list(f.value))
        return names

    @property
    def schemata(self) -> set[SchemaFilter]:
        """
        The current schema filters
        """
        return {f for f in self.filters if isinstance(f, SchemaFilter)}

    @property
    def schemata_names(self) -> set[str]:
        """
        The names of the current filtered schemas
        """
        names = set()
        for f in self.schemata:
            names.update(ensure_list([s.name for s in f.schemata]))
        return names

    @property
    def countries(self) -> set[str]:
        """
        The current filtered countries
        """
        names = set()
        for f in self.properties:
            if f.key == "country":
                names.update(ensure_list(f.value))
        return names

    @property
    def reversed(self) -> set[ReverseFilter]:
        """
        The current reverse lookup filters
        """
        return {f for f in self.filters if isinstance(f, ReverseFilter)}

    @property
    def properties(self) -> set[PropertyFilter]:
        """
        The current property lookup filters
        """
        return {f for f in self.filters if isinstance(f, PropertyFilter)}

    def discard(self, f_cls: F) -> None:
        filters = list(self.filters)
        for f in filters:
            if isinstance(f, f_cls):
                self.filters.discard(f)

    def to_dict(self) -> dict[str, Any]:
        """
        Dictionary representation of the current object

        Example:
            ```python
            q = Query().where(dataset__in=["d1", "d2"])
            assert q.to_dict() == {"dataset__in": {"d1", "d2"}}
            q = q.where(schema="Event").where(schema__in=["Person", "Organization"])
            assert q.to_dict() == {
                    "dataset__in": {"d1", "d2"},
                    "schema": "Event",
                    "schema__in": {"Organization", "Person"},
                }
            ```
        """
        data = self.lookups
        if self.sort:
            data["order_by"] = self.sort.serialize()
        if self.slice:
            data["limit"] = self.limit
            data["offset"] = self.offset
        if self.aggregations:
            data["aggregations"] = self.get_aggregator().to_dict()
        return data

    def where(self, **lookup: Any) -> Q:
        """
        Add another lookup to the current `Query` instance.

        Example:
            ```python
            q = Query().where(dataset="my_dataset")
            q = q.where(schema="Payment")
            q = q.where(date__gte="2024-10", date__lt="2024-11")
            q = q.order_by("amountEur", ascending=False)
            ```

        Args:
            **lookup: A dataset lookup `dataset="my_dataset"`
            **lookup: A schema lookup `schema="Person"`
            **lookup: `schema_include_descendants=True`: Include schema descendants for
                given schema lookup
            **lookup: `schema_include_matchable=True`: Include matchable schema for
                given schema lookup
            **lookup: A property=value lookup (with optional comparators):
                `name__startswith="Ja"`

        Returns:
            The updated `Query` instance
        """
        schema_include_descendants = lookup.pop("schema_include_descendants", False)
        schema_include_matchable = lookup.pop("schema_include_matchable", False)
        prop = lookup.pop("prop", None)
        value = lookup.pop("value", None)
        comparator = lookup.pop("comparator", None)
        if prop is not None:
            if value is None:
                raise ValidationError("No lookup value specified")
            f = PropertyFilter(prop, value, comparator)
            self.filters.discard(f)  # replace existing property filter with updated one
            self.filters.add(f)

        properties: dict[str, Any] = {}
        for key, value in lookup.items():
            meta = False
            for f_key, f in FILTERS.items():
                if key.startswith(f_key):
                    if value is None:
                        self.discard(f)
                    else:
                        key, comparator = parse_comparator(key)
                        kwargs = {}
                        if key == "schema":
                            kwargs = {
                                "schema_include_matchable": schema_include_matchable,
                                "schema_include_descendants": schema_include_descendants,
                            }
                        self.filters.add(f(value, comparator, **kwargs))
                    meta = True
                    break
            if not meta:
                properties[key] = value

        # parse arbitrary `date__gte=2023` stuff
        for key, val in properties.items():
            for prop, value, comparator in parse_unknown_filters((key, val)):
                f = PropertyFilter(prop, value, comparator)
                self.filters.discard(
                    f
                )  # replace existing property filter with updated one
                self.filters.add(f)

        return self._chain()

    def order_by(self, *values: Iterable[str], ascending: bool | None = True) -> Q:
        """
        Add or update the current sorting.

        Args:
            *values: Fields to order by
            ascending: Ascending or descending

        Returns:
            The updated `Query` instance.
        """
        self.sort = Sort(values=values, ascending=ascending)
        return self._chain()

    def aggregate(
        self,
        func: Aggregations,
        *props: Properties,
        groups: Properties | list[Properties] | None = None,
    ) -> Q:
        for prop in props:
            self.aggregations.add(
                Aggregation(func=func, prop=prop, group_props=ensure_list(groups))
            )
        return self._chain()

    def get_aggregator(self) -> Aggregator:
        return Aggregator(aggregations=self.aggregations)

    def apply(self, proxy: CE) -> bool:
        """
        Test if a proxy matches the current `Query` instance.
        """
        if not self.filters:
            return True
        return all(f.apply(proxy) for f in self.filters)

    def apply_iter(self, proxies: CEGenerator) -> CEGenerator:
        """
        Apply the current `Query` instance to a generator of proxies and return
        a generator of filtered proxies

        Example:
            ```python
            proxies = [...]
            q = Query().where(dataset="my_dataset", schema="Company")
            for proxy in q.apply_iter(proxies):
                assert proxy.schema.name == "Company"
            ```

        Yields:
            A generator of `nomenklatura.entity.CompositeEntity`
        """
        if not self:
            yield from proxies
            return

        proxies = (p for p in proxies if self.apply(p))
        if self.sort:
            proxies = self.sort.apply_iter(proxies)
        if self.slice:
            proxies = islice(
                proxies, self.slice.start, self.slice.stop, self.slice.step
            )
        if self.aggregations:
            self.aggregator = self.get_aggregator()
            proxies = self.aggregator.apply(proxies)
        yield from proxies



import os
from collections import defaultdict
from decimal import Decimal

from anystore.util import clean_dict
from nomenklatura import store as nk
from nomenklatura.dataset import DS
from sqlalchemy import select

from ftmq.aggregations import AggregatorResult
from ftmq.enums import Fields
from ftmq.exceptions import ValidationError
from ftmq.model.coverage import Collector, DatasetStats
from ftmq.model.dataset import Catalog
from ftmq.query import Q, Query
from ftmq.store.base import Store, View
from ftmq.types import CEGenerator
from ftmq.util import to_numeric

MAX_SQL_AGG_GROUPS = int(os.environ.get("MAX_SQL_AGG_GROUPS", 10))


def clean_agg_value(value: str | Decimal) -> str | float | int | None:
    if isinstance(value, Decimal):
        return to_numeric(value)
    return value


class SQLQueryView(View, nk.sql.SQLView):
    def ensure_scoped_query(self, query: Q) -> Q:
        if not query.datasets:
            return query.where(dataset__in=self.dataset_names)
        if query.dataset_names - self.dataset_names:
            raise ValidationError("Query datasets outside view scope")
        return query

    def entities(self, query: Q | None = None) -> CEGenerator:
        if query:
            query = self.ensure_scoped_query(query)
            yield from self.store._iterate(query.sql.statements)
        else:
            view = self.store.view(self.scope)
            yield from view.entities()

    def stats(self, query: Q | None = None) -> DatasetStats:
        query = self.ensure_scoped_query(query or Query())
        key = f"stats-{hash(query)}"
        if key in self._cache:
            return self._cache[key]

        c = Collector()
        for schema, count in self.store._execute(query.sql.things, stream=False):
            c.things[schema] = count
        for schema, count in self.store._execute(query.sql.intervals, stream=False):
            c.intervals[schema] = count
        for country, count in self.store._execute(
            query.sql.things_countries, stream=False
        ):
            if country is not None:
                c.things_countries[country] = count
        for country, count in self.store._execute(
            query.sql.intervals_countries, stream=False
        ):
            if country is not None:
                c.intervals_countries[country] = count

        stats = c.export()
        for start, end in self.store._execute(query.sql.date_range, stream=False):
            if start:
                stats.coverage.start = start
            if end:
                stats.coverage.end = end
            break

        stats.entity_count = self.count(query)
        self._cache[key] = stats
        return stats

    def count(self, query: Q | None = None) -> int:
        if query is not None:
            for res in self.store._execute(query.sql.count, stream=False):
                for count in res:
                    return count
        return 0

    def aggregations(self, query: Q) -> AggregatorResult | None:
        if not query.aggregations:
            return
        query = self.ensure_scoped_query(query)
        key = f"agg-{hash(query)}"
        if key in self._cache:
            return self._cache[key]
        res: AggregatorResult = defaultdict(dict)

        for prop, func, value in self.store._execute(
            query.sql.aggregations, stream=False
        ):
            res[func][prop] = clean_agg_value(value)

        if query.sql.group_props:
            res["groups"] = defaultdict(lambda: defaultdict(lambda: defaultdict(dict)))
            for prop in query.sql.group_props:
                if prop == Fields.year:
                    start, end = self.stats(query).coverage.years
                    if start or end:
                        groups = range(start or end, (end or start) + 1)
                    else:
                        groups = []
                else:
                    groups = [
                        r[0]
                        for r in self.store._execute(
                            query.sql.get_group_counts(prop, limit=MAX_SQL_AGG_GROUPS),
                            stream=False,
                        )
                    ]
                for group in groups:
                    for agg_prop, func, value in self.store._execute(
                        query.sql.get_group_aggregations(prop, group), stream=False
                    ):
                        res["groups"][prop][func][agg_prop][group] = clean_agg_value(
                            value
                        )
        res = clean_dict(res)
        self._cache[key] = res
        return res


class SQLStore(Store, nk.SQLStore):
    def get_catalog(self) -> Catalog:
        q = select(self.table.c.dataset).distinct()
        names: set[str] = set()
        for row in self._execute(q, stream=False):
            names.add(row[0])
        return Catalog.from_names(names)

    def query(self, scope: DS | None = None, external: bool = False) -> SQLQueryView:
        scope = scope or self.dataset
        return SQLQueryView(self, scope, external=external)



from nomenklatura import store as nk
from nomenklatura.dataset import DS

from ftmq.model.dataset import Catalog
from ftmq.store.base import Store, View


class MemoryQueryView(View, nk.memory.MemoryView):
    pass


class MemoryStore(Store, nk.SimpleMemoryStore):
    def get_catalog(self) -> Catalog:
        return Catalog.from_names(self.entities.keys())

    def query(self, scope: DS | None = None, external: bool = False) -> MemoryQueryView:
        scope = scope or self.dataset
        return MemoryQueryView(self, scope, external=external)



from ftmq.io import smart_read_proxies, smart_stream_proxies, smart_write_proxies
from ftmq.query import Query
from ftmq.util import make_proxy

__version__ = "0.8.7"
__all__ = [
    "smart_read_proxies",
    "smart_stream_proxies",
    "smart_write_proxies",
    "Query",
    "make_proxy",
]



from nomenklatura.dataset import DS
from nomenklatura.store import level

from ftmq.model import Catalog
from ftmq.store.base import Store, View


class LevelDBQueryView(View, level.LevelDBView):
    pass


class LevelDBStore(Store, level.LevelDBStore):
    def get_catalog(self) -> Catalog:
        names: set[str] = set()
        with self.db.iterator(prefix=b"e:", include_value=False) as it:
            for k in it:
                _, _, dataset = k.decode("utf-8").split(":", 2)
                names.add(dataset)
        return Catalog.from_names(names)

    def query(
        self, scope: DS | None = None, external: bool = False
    ) -> LevelDBQueryView:
        scope = scope or self.dataset
        return LevelDBQueryView(self, scope, external=external)



from typing import Self

from nomenklatura.dataset import DS
from nomenklatura.resolver import Resolver

from ftmq.aleph import AlephStore as _AlephStore
from ftmq.aleph import AlephView, parse_uri
from ftmq.model.dataset import Catalog, Dataset
from ftmq.store.base import Store, View
from ftmq.util import DefaultDataset


class AlephQueryView(View, AlephView):
    pass


class AlephStore(Store, _AlephStore):
    def get_catalog(self) -> Catalog:
        # FIXME
        # api.filter_collections("*")
        return Catalog.from_names(DefaultDataset.leaf_names)

    def query(self, scope: DS | None = None, external: bool = False) -> AlephQueryView:
        scope = scope or self.dataset
        return AlephQueryView(self, scope, external=external)

    @classmethod
    def from_uri(
        cls,
        uri: str,
        dataset: Dataset | str | None = None,
        catalog: Catalog | None = None,
        linker: Resolver | None = None,
    ) -> Self:
        host, api_key, foreign_id = parse_uri(uri)
        if dataset is None and foreign_id is not None:
            dataset = foreign_id
        if dataset is not None:
            if isinstance(dataset, str):
                dataset = Dataset(name=dataset)

        return cls(catalog, dataset, linker=linker, host=host, api_key=api_key)



from typing import Generator

from nomenklatura.judgement import Judgement
from nomenklatura.resolver import Resolver
from sqlalchemy import desc, or_, select


def get_similar(
    entity_id: str, resolver: Resolver, limit: int | None = None
) -> Generator[tuple[str, float], None, None]:
    """
    Get similar entity ids for given id with NO_JUDGEMENT from resolver
    """

    t = resolver._table
    stmt = select(t.c.target, t.c.source, t.c.score)
    stmt = stmt.where(or_(t.c.source == entity_id, t.c.target == entity_id))
    stmt = stmt.where(t.c.judgement == Judgement.NO_JUDGEMENT.value)
    stmt = stmt.order_by(desc(t.c.score))
    if limit:
        stmt = stmt.limit(limit)
    resolver.begin()
    cursor = resolver._get_connection().execute(stmt)
    while batch := cursor.fetchmany(10000):
        for target, source, score in batch:
            if target != entity_id:
                yield target, score
            if source != entity_id:
                yield source, score
    resolver.commit()



from functools import cache
from typing import Generator, Iterable

from nomenklatura import CompositeEntity
from nomenklatura import store as nk
from nomenklatura.db import get_engine
from nomenklatura.resolver import Resolver

from ftmq.aggregations import AggregatorResult
from ftmq.logging import get_logger
from ftmq.model.coverage import Collector, DatasetStats
from ftmq.model.dataset import C, Dataset
from ftmq.query import Q
from ftmq.similar import get_similar
from ftmq.types import CE, CEGenerator
from ftmq.util import DefaultDataset, ensure_dataset, make_dataset

log = get_logger(__name__)


@cache
def get_resolver(uri: str | None = None) -> Resolver[CompositeEntity]:
    return Resolver.make_default(get_engine(uri))


class Store(nk.Store):
    """
    Feature add-ons to `nomenklatura.store.Store`
    """

    def __init__(
        self,
        catalog: C | None = None,
        dataset: Dataset | str | None = None,
        linker: Resolver | None = None,
        **kwargs,
    ) -> None:
        """
        Initialize a store. This should be called via
        [`get_store`][ftmq.store.get_store]

        Args:
            catalog: A `ftmq.model.Catalog` instance to limit the scope to
            dataset: A `ftmq.model.Dataset` instance to limit the scope to
            linker: A `nomenklatura.Resolver` instance with linked / deduped data
        """
        if dataset is not None:
            if isinstance(dataset, str):
                dataset = Dataset(name=dataset)
            dataset = make_dataset(dataset.name)
        elif catalog is not None:
            dataset = catalog.get_scope()
        else:
            dataset = DefaultDataset
        linker = linker or get_resolver()
        super().__init__(dataset=dataset, linker=linker, **kwargs)
        # implicit set all datasets as default store scope:
        if dataset == DefaultDataset:
            self.dataset = self.get_catalog().get_scope()

    def get_catalog(self) -> C:
        """
        Return implicit `Catalog` computed from current datasets in store
        """
        raise NotImplementedError

    def iterate(self, dataset: str | Dataset | None = None) -> CEGenerator:
        """
        Iterate all the entities, optional filter for a dataset.

        Args:
            dataset: `Dataset` instance or name to limit scope to

        Yields:
            Generator of `nomenklatura.entity.CompositeEntity`
        """
        dataset = ensure_dataset(dataset)
        if dataset is not None:
            view = self.view(dataset)
        else:
            catalog = self.get_catalog()
            view = self.view(catalog.get_scope())
        yield from view.entities()


class View(nk.base.View):
    """
    Feature add-ons to `nomenklatura.store.base.View`
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._cache = {}

    def entities(self, query: Q | None = None) -> CEGenerator:
        """
        Get the entities of a store, optionally filtered by a
        [`Query`][ftmq.Query] object.

        Args:
            query: The Query filter object

        Yields:
            Generator of `nomenklatura.entity.CompositeEntity`
        """
        view = self.store.view(self.scope)
        if query:
            yield from query.apply_iter(view.entities())
        else:
            yield from view.entities()

    def get_adjacents(
        self, proxies: Iterable[CE], inverted: bool | None = False
    ) -> set[CE]:
        seen: set[CE] = set()
        for proxy in proxies:
            for _, adjacent in self.get_adjacent(proxy, inverted=inverted):
                if adjacent.id not in seen:
                    seen.add(adjacent)
        return seen

    def stats(self, query: Q | None = None) -> DatasetStats:
        key = f"stats-{hash(query)}"
        if key in self._cache:
            return self._cache[key]
        c = Collector()
        cov = c.collect_many(self.entities(query))
        self._cache[key] = cov
        return cov

    def count(self, query: Q | None = None) -> int:
        return self.stats(query).entity_count or 0

    def aggregations(self, query: Q) -> AggregatorResult | None:
        if not query.aggregations:
            return
        key = f"agg-{hash(query)}"
        if key in self._cache:
            return self._cache[key]
        _ = [x for x in self.entities(query)]
        res = dict(query.aggregator.result)
        self._cache[key] = res
        return res

    def similar(
        self, entity_id: str, limit: int | None = None
    ) -> Generator[tuple[CE, float], None, None]:
        for candidate_id, score in get_similar(entity_id, self.store.linker, limit):
            yield self.get_entity(candidate_id), score



from ftmq.model.coverage import DatasetStats
from ftmq.model.dataset import Catalog, Dataset, Maintainer, Publisher, Resource
from ftmq.model.proxy import Entity

__all__ = [
    "Catalog",
    "DatasetStats",
    "Dataset",
    "Entity",
    "Maintainer",
    "Publisher",
    "Resource",
]



from functools import cache
from pathlib import Path
from typing import TypeVar
from urllib.parse import urlparse

from nomenklatura import Resolver, settings
from nomenklatura.db import get_metadata

from ftmq.model.dataset import Catalog, Dataset
from ftmq.store.aleph import AlephStore
from ftmq.store.base import Store, View
from ftmq.store.memory import MemoryStore
from ftmq.store.sql import SQLStore
from ftmq.types import PathLike

S = TypeVar("S", bound=Store)


@cache
def get_store(
    uri: PathLike | None = settings.DB_URL,
    catalog: Catalog | None = None,
    dataset: Dataset | str | None = None,
    linker: Resolver | None = None,
) -> Store:
    """
    Get an initialized [Store][ftmq.store.base.Store]. The backend is inferred
    by the scheme of the store uri.

    Example:
        ```python
        from ftmq.store import get_store

        # an in-memory store:
        get_store("memory://")

        # a leveldb store:
        get_store("leveldb:///var/lib/data")

        # a redis (or kvrocks) store:
        get_store("redis://localhost")

        # a sqlite store
        get_store("sqlite:///data/followthemoney.db")
        ```

    Args:
        uri: The store backend uri
        catalog: A `ftmq.model.Catalog` instance to limit the scope to
        dataset: A `ftmq.model.Dataset` instance to limit the scope to
        linker: A `nomenklatura.Resolver` instance with linked / deduped data

    Returns:
        The initialized store. This is a cached object.
    """
    if isinstance(dataset, str):
        dataset = Dataset(name=dataset)
    uri = str(uri)
    parsed = urlparse(uri)
    if parsed.scheme == "memory":
        return MemoryStore(catalog, dataset, linker=linker)
    if parsed.scheme == "leveldb":
        path = uri.replace("leveldb://", "")
        path = Path(path).absolute()
        try:
            from ftmq.store.level import LevelDBStore

            return LevelDBStore(catalog, dataset, path=path, linker=linker)
        except ImportError:
            raise ImportError("Can not load LevelDBStore. Install `plyvel`")
    if parsed.scheme == "redis":
        try:
            from ftmq.store.redis import RedisStore

            return RedisStore(catalog, dataset, linker=linker)
        except ImportError:
            raise ImportError("Can not load RedisStore. Install `redis`")
    if parsed.scheme == "clickhouse":
        try:
            from ftm_columnstore import get_store as get_cstore

            return get_cstore(catalog, dataset, linker=linker)
        except ImportError:
            raise ImportError("Can not load ClickhouseStore. Install `ftm-columnstore`")
    if "sql" in parsed.scheme:
        get_metadata.cache_clear()
        return SQLStore(catalog, dataset, uri=uri, linker=linker)
    if "aleph" in parsed.scheme:
        return AlephStore.from_uri(uri, catalog=catalog, dataset=dataset, linker=linker)
    raise NotImplementedError(uri)


__all__ = ["get_store", "S", "Store", "View", "MemoryStore", "SQLStore", "AlephStore"]



from collections import Counter
from typing import Any

from followthemoney import model

from ftmq.enums import Properties
from ftmq.model.mixins import BaseModel
from ftmq.types import CE, CEGenerator, Frequencies
from ftmq.util import get_country_name, get_year_from_iso


class Schema(BaseModel):
    name: str
    count: int
    label: str
    plural: str

    def __init__(self, **data):
        schema = model.get(data["name"])
        data["label"] = schema.label
        data["plural"] = schema.plural
        super().__init__(**data)


class Country(BaseModel):
    code: str
    count: int
    label: str | None = None

    def __init__(self, **data):
        data["label"] = get_country_name(data["code"])
        super().__init__(**data)


class Schemata(BaseModel):
    total: int = 0
    countries: list[Country] | None = []
    schemata: list[Schema] | None = []


class Coverage(BaseModel):
    start: str | None = None
    end: str | None = None
    frequency: Frequencies | None = "unknown"
    countries: list[str] | None = []
    schedule: str | None = None

    @property
    def years(self) -> tuple[int | None, int | None]:
        """
        Return min / max year extend
        """
        return get_year_from_iso(self.start), get_year_from_iso(self.end)


class DatasetStats(BaseModel):
    coverage: Coverage
    things: Schemata
    intervals: Schemata
    entity_count: int | None = 0


class Collector:
    def __init__(self):
        self.things = Counter()
        self.things_countries = Counter()
        self.intervals = Counter()
        self.intervals_countries = Counter()
        self.start = set()
        self.end = set()

    def collect(self, proxy: CE) -> None:
        if proxy.schema.is_a("Thing"):
            self.things[proxy.schema.name] += 1
            for country in proxy.countries:
                self.things_countries[country] += 1
        else:
            self.intervals[proxy.schema.name] += 1
            for country in proxy.countries:
                self.intervals_countries[country] += 1
        self.start.update(proxy.get(Properties.startDate, quiet=True))
        self.start.update(proxy.get(Properties.date, quiet=True))
        self.end.update(proxy.get(Properties.endDate, quiet=True))
        self.end.update(proxy.get(Properties.date, quiet=True))

    def export(self) -> DatasetStats:
        start = min(self.start) if self.start else None
        end = max(self.end) if self.end else None
        countries = set(self.things_countries.keys()) | set(
            self.intervals_countries.keys()
        )
        coverage = Coverage(start=start, end=end, countries=countries)
        things = Schemata(
            schemata=[Schema(name=k, count=v) for k, v in self.things.items()],
            countries=[
                Country(code=k, count=v) for k, v in self.things_countries.items()
            ],
            total=self.things.total(),
        )
        intervals = Schemata(
            schemata=[Schema(name=k, count=v) for k, v in self.intervals.items()],
            countries=[
                Country(code=k, count=v) for k, v in self.intervals_countries.items()
            ],
            total=self.intervals.total(),
        )
        return DatasetStats(
            coverage=coverage,
            things=things,
            intervals=intervals,
            entity_count=things.total + intervals.total,
        )

    def to_dict(self) -> dict[str, Any]:
        data = self.export()
        return data.model_dump()

    def apply(self, proxies: CEGenerator) -> CEGenerator:
        """
        Generate coverage from an input stream of proxies
        This returns a generator again, so actual collection of coverage stats
        will happen if the actual generator is executed
        """
        for proxy in proxies:
            self.collect(proxy)
            yield proxy

    def collect_many(self, proxies: CEGenerator) -> DatasetStats:
        for proxy in proxies:
            self.collect(proxy)
        return self.export()



from anystore.mixins import BaseModel as _BaseModel
from pydantic import Field


class BaseModel(_BaseModel):
    from_uri: str | None = Field(default=None, exclude=True)

    def __init__(self, **data):
        from_uri = data.pop("from_uri", None)
        if from_uri is not None:
            data = self._from_uri(from_uri, **data).model_dump()
            if "uri" in self.model_fields:
                data["uri"] = data.get("uri") or from_uri
        super().__init__(**data)



from typing import Any, Iterable, Self, Sequence, TypeAlias, TypeVar, Union

from followthemoney.types import registry
from nomenklatura.publish.names import pick_caption
from pydantic import BaseModel, ConfigDict, Field, model_validator

from ftmq.types import CE
from ftmq.util import make_proxy, must_str

EntityProp = TypeVar("EntityProp", bound="Entity")
Properties: TypeAlias = dict[str, Sequence[Union[str, EntityProp]]]


class Entity(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    id: str = Field(..., examples=["NK-A7z...."])
    caption: str = Field(..., examples=["Jane Doe"])
    schema_: str = Field(..., examples=["LegalEntity"], alias="schema")
    properties: Properties = Field(..., examples=[{"name": ["Jane Doe"]}])
    datasets: list[str] = Field([], examples=[["us_ofac_sdn"]])
    referents: list[str] = Field([], examples=[["ofac-1234"]])

    @classmethod
    def from_proxy(cls, entity: CE, adjacents: Iterable[CE] | None = None) -> Self:
        properties = dict(entity.properties)
        if adjacents:
            adjacents_: dict[str, Entity] = {
                must_str(e.id): Entity.from_proxy(e) for e in adjacents
            }
            for prop in entity.iterprops():
                if prop.type == registry.entity:
                    properties[prop.name] = [
                        adjacents_.get(i, i) for i in entity.get(prop)
                    ]
        return cls(
            id=must_str(entity.id),
            caption=entity.caption,
            schema=entity.schema.name,
            properties=properties,
            datasets=list(entity.datasets),
            referents=list(entity.referents),
        )

    def to_proxy(self) -> CE:
        return make_proxy(self.model_dump(by_alias=True))

    @model_validator(mode="before")
    @classmethod
    def get_caption(cls, data: Any) -> Any:
        if data.get("caption") is None:
            data["caption"] = pick_caption(make_proxy(data))
        return data



"""
Overwrite `ftm aggregate` with the possibility to merge via common parent
schemata.
"""

from typing import Iterable

from anystore.io import logged_items
from followthemoney import model
from followthemoney.exc import InvalidData
from followthemoney.schema import Schema

from ftmq.enums import Schemata
from ftmq.types import CEGenerator, Proxy
from ftmq.util import make_proxy

SCHEMATA: dict[Schemata, int] = {s: len(model.get(s).extends) for s in Schemata}


def extends(s: Schema) -> set[Schema]:
    schemata: set[Schema] = set()
    for schema in s.extends:
        schemata.add(schema)
        for e_schema in schema.extends:
            schemata.add(e_schema)
            schemata.update(extends(e_schema))
    return schemata


def common_ancestor(s1: Schema, s2: Schema) -> Schema:
    ancestors = extends(s1) & extends(s2)
    for schema in sorted(ancestors, key=lambda x: SCHEMATA[x.name], reverse=True):
        return schema
    raise InvalidData(f"No common ancestors: {s1}, {s2}")


def merge(p1: Proxy, p2: Proxy, downgrade: bool | None = False) -> Proxy:
    try:
        p1 = p1.merge(p2)
        p1.schema = model.common_schema(p1.schema, p2.schema)
        return p1
    except InvalidData as e:
        if downgrade:
            # try common schemata, this will probably "downgrade" entities
            # as in, losing some schema specific properties
            schema = common_ancestor(p1.schema, p2.schema)
            p1_data = p1.to_full_dict()
            p1_data["schema"] = schema.name
            p2_data = p2.to_full_dict()
            p2_data["schema"] = schema.name
            p1 = make_proxy(p1_data)
            p2 = make_proxy(p2_data)
            return p1.merge(p2)

        raise e


def aggregate(proxies: Iterable[Proxy], downgrade: bool | None = False) -> CEGenerator:
    buffer: dict[str, Proxy] = {}
    for proxy in logged_items(proxies, "Aggregate", item_name="Proxy"):
        if proxy.id in buffer:
            buffer[proxy.id] = merge(buffer[proxy.id], proxy, downgrade)
        else:
            buffer[proxy.id] = proxy
    yield from buffer.values()



from datetime import datetime
from typing import Iterable, Literal, Self, TypeVar

from anystore.io import logged_items
from nomenklatura.dataset.dataset import Dataset as NKDataset
from normality import slugify
from pydantic import AnyUrl, HttpUrl
from rigour.mime.types import FTM

from ftmq.enums import Categories, Frequencies
from ftmq.model.coverage import Coverage, DatasetStats, Schemata
from ftmq.model.mixins import BaseModel
from ftmq.types import CEGenerator, SDict
from ftmq.util import make_dataset

Frequencies = Literal[tuple(Frequencies)]
Categories = Literal[tuple(Categories)]
ContentType = Literal["documents", "structured", "mixed"]

C = TypeVar("C", bound="Catalog")
DS = TypeVar("DS", bound="Dataset")


class Publisher(BaseModel):
    name: str
    url: HttpUrl
    description: str | None = None
    country: str | None = None
    country_label: str | None = None
    official: bool = False
    logo_url: HttpUrl | None = None


class Resource(BaseModel):
    name: str
    url: AnyUrl
    title: str | None = None
    checksum: str | None = None
    timestamp: datetime | None = None
    mime_type: str | None = None
    mime_type_label: str | None = None
    size: int | None = 0


class Maintainer(BaseModel):
    """
    this is our own addition
    """

    name: str
    description: str | None = None
    country: str | None = None
    country_label: str | None = None
    url: HttpUrl | None = None
    logo_url: HttpUrl | None = None


class Dataset(BaseModel):
    # nk props
    name: str
    prefix: str | None = None
    title: str | None = None
    license: str | None = None
    summary: str | None = None
    description: str | None = None
    url: HttpUrl | None = None
    updated_at: datetime | None = None
    version: str | None = None
    category: Categories | None = None
    publisher: Publisher | None = None
    coverage: Coverage | None = None
    things: Schemata | None = None
    intervals: Schemata | None = None
    entity_count: int | None = 0
    resources: list[Resource] | None = []
    index_url: AnyUrl | None = None

    # own addition / aleph
    catalog: str | None = None
    countries: list[str] | None = []
    info_url: HttpUrl | None = None
    data_url: HttpUrl | None = None
    aleph_url: HttpUrl | None = None
    tags: list[str] | None = []
    content_type: ContentType | None = "structured"
    total_file_size: int | None = 0

    git_repo: AnyUrl | None = None
    uri: str | None = None
    maintainer: Maintainer | None = None

    def __init__(self, **data):
        data["updated_at"] = data.get("updated_at") or datetime.utcnow().replace(
            microsecond=0
        )
        super().__init__(**data)
        self.prefix = self.prefix or data.get("prefix") or slugify(self.name)
        self.coverage = self.coverage or Coverage()
        self.title = self.title or self.name.title()

    def iterate(self) -> CEGenerator:
        from ftmq.io import smart_read_proxies  # FIXME

        for resource in self.resources:
            if resource.mime_type == FTM:
                yield from logged_items(
                    smart_read_proxies(resource.url),
                    "Read",
                    1_000,
                    uri=resource.url,
                    item_name="Proxy",
                    dataset=self.name,
                )

    def apply_stats(self, stats: DatasetStats) -> None:
        self.entity_count = stats.entity_count
        self.coverage = stats.coverage
        self.things = stats.things
        self.intervals = stats.intervals


def ensure_dataset(data: SDict | Dataset) -> Dataset:
    if isinstance(data, Dataset):
        return data
    return Dataset(**data)


class Catalog(BaseModel):
    name: str | None = "default"
    title: str | None = "Catalog"
    datasets: list[Dataset] | None = []
    updated_at: datetime | None = None
    description: str | None = None
    maintainer: Maintainer | None = None
    publisher: Publisher | None = None
    url: HttpUrl | None = None
    uri: str | None = None
    logo_url: HttpUrl | None = None
    git_repo: AnyUrl | None = None

    def __init__(self, **data):
        if "name" not in data:
            data["name"] = "Catalog"
        data["datasets"] = [ensure_dataset(d) for d in data.get("datasets", [])]
        super().__init__(**data)

    def get(self, name: str) -> Dataset | None:
        for dataset in self.datasets:
            if dataset.name == name:
                return dataset

    def get_scope(self) -> NKDataset:
        # FIXME clarify
        ds = NKDataset(
            {
                "name": slugify(self.name),
                "title": self.name.title(),
                "children": self.names,
            },
        )
        ds.children = {make_dataset(n) for n in self.names}
        return ds

    def iterate(self) -> CEGenerator:
        for dataset in self.datasets:
            yield from dataset.iterate()

    @property
    def names(self) -> set[str]:
        names = set()
        for dataset in self.datasets:
            names.add(dataset.name)
        return names

    @classmethod
    def from_names(cls, names: Iterable[str]) -> Self:
        return cls(datasets=[Dataset(name=n) for n in names])
